from __future__ import annotations
import re
import os
import os.path
import sys
import uuid
import psutil
import shutil
import operator
import subprocess

import winreg
import win32com.client
import win32com.client.gencache
import win32event
import win32api
import win32com
import win32con
import win32process
import pythoncom
import warnings

from typing import Union, Optional, Any
from typing_extensions import TypeAlias

Version: TypeAlias = tuple[int, int]


class E3Warning(UserWarning):
    pass


def get_version_from_executable(executable_path: str) -> Version:
    file_info = win32api.GetFileVersionInfo(executable_path, "\\")
    ms = file_info["FileVersionMS"]
    return (win32api.HIWORD(ms), win32api.LOWORD(ms))


def get_executable_from_pid(pid: int) -> str:
    process_handle = win32api.OpenProcess(win32con.PROCESS_QUERY_INFORMATION | win32con.PROCESS_VM_READ, False, pid)
    executable_path = win32process.GetModuleFileNameEx(process_handle, 0)
    win32api.CloseHandle(process_handle)
    return str(executable_path)


def get_commandline_args_from_pid(pid: int) -> Optional[str]:
    strComputer = "."
    obj_wmi_service = win32com.client.Dispatch("WbemScripting.SWbemLocator")
    obj_swbem_services = obj_wmi_service.ConnectServer(strComputer, "root/cimv2")

    proc_items = list(obj_swbem_services.ExecQuery(f"Select * from Win32_Process where ProcessId='{int(pid)}'"))
    if len(proc_items) == 0:
        raise RuntimeError(f"Invalid process ID '{pid}'")

    obj_item = proc_items[0]
    assert obj_item.ProcessId == pid
    return str(obj_item.CommandLine)


def get_installed() -> list[tuple[Version, str]]:
    """
    Returns a list of all installed E3.series. The installed versions are read from the registry. Only existing Installations are returned.

    :return: a list of tuples. Every tuple corresponds to the schema (<version>, <executable path>), e.g. ((25, 0), 'C:\\Program Files\\Zuken\\E3.series_2025\\E3.series.exe'). The list is always sorted by version with the latest version at the end.
    """
    with winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE) as registry:
        versions: list[str] = []
        try:
            with winreg.OpenKey(registry, "SOFTWARE\\Zuken\\E3.series") as root_key:
                i = 0
                while True:
                    try:
                        child_key = winreg.EnumKey(root_key, i)
                        versions.append(child_key)
                        i += 1
                    except OSError:
                        break
        except FileNotFoundError:
            pass

        version_and_path: list[tuple[Version, str]] = []
        for version in versions:
            with winreg.OpenKey(registry, f"SOFTWARE\\Zuken\\E3.series\\{version}") as key:
                try:
                    value, type = winreg.QueryValueEx(key, "InstallationPath")
                    if type != winreg.REG_SZ:
                        continue

                    executable_path = os.path.join(value, "E3.series.exe")
                    if not os.path.isfile(executable_path):
                        continue

                    # We read the product version, because it includes the minor version.
                    # The registry folder does not include the service pack and patch version
                    file_version = get_version_from_executable(executable_path)
                    version_and_path.append((file_version, executable_path))
                except FileNotFoundError:
                    pass

    version_and_path.sort(key=operator.itemgetter(0))
    return version_and_path


def get_registered_typelibs() -> list[tuple[Version, str]]:
    """
    Returns registered E3.series COM typelibs. Only existing typelibs are returned.
    """
    with winreg.ConnectRegistry(None, winreg.HKEY_CLASSES_ROOT) as registry:
        hex_versions: list[str] = []
        try:
            with winreg.OpenKey(registry, "TypeLib\\{59B53E00-679B-41ED-9D11-766E27B6A7A4}") as root_key:
                i = 0
                while True:
                    try:
                        child_key = winreg.EnumKey(root_key, i)
                        hex_versions.append(child_key)
                        i += 1
                    except OSError:
                        break

        except FileNotFoundError:
            pass

        version_and_path: list[tuple[Version, str]] = []
        for hex_version in hex_versions:
            for bitness in ["win32", "win64"]:
                try:
                    with winreg.OpenKey(registry, f"TypeLib\\{{59B53E00-679B-41ED-9D11-766E27B6A7A4}}\\{hex_version}\\0\\{bitness}") as key:
                        tlb_path, type = winreg.QueryValueEx(key, "")
                        if type != winreg.REG_SZ:
                            continue

                        hex_pieces = hex_version.split(".")
                        assert len(hex_pieces) == 2
                        version = (int(hex_pieces[0], 16), int(hex_pieces[1], 16))

                        if not os.path.isfile(tlb_path):
                            warnings.warn(
                                f'E3.series version {version[0]}.{version[1]} was not uninstalled correctly. The typelib is still registered, but the file "{tlb_path}" does not exist anymore.',
                                E3Warning,
                            )
                            continue

                        version_and_path.append((version, tlb_path))
                except FileNotFoundError:
                    pass

    version_and_path.sort(key=operator.itemgetter(0))
    return version_and_path


class StartArguments:
    def __init__(self) -> None:
        self.filename: Optional[str] = None
        self.cable: bool = False
        self.compare_config_file: Optional[str] = None
        self.compare_new_file: Optional[str] = None
        self.compare_old_file: Optional[str] = None
        self.dbe: bool = False
        self.demo: bool = False
        self.dist_design: bool = False
        self.economy: bool = False
        self.fluid: bool = False
        self.formboard: bool = False
        self.functional_design: bool = False
        self.inprocreg_only: bool = False
        self.level: bool = False
        self.logic: bool = False
        self.mu_create: bool = False
        self.multiuser: bool = False
        self.mu_open: Union[bool, str, None] = False
        self.new: bool = False
        self.no_cgm: bool = False
        self.no_dbe: bool = False
        self.no_export_exf: bool = False
        self.no_import_ruplan: bool = False
        self.no_import_step: bool = False
        self.no_mil_standard: bool = False
        self.no_panel: bool = False
        self.no_pdf: bool = False
        self.no_plugin: bool = False
        self.no_splash: bool = False
        self.no_wire: bool = False
        self.no_xvl: bool = False
        self.plus: bool = False
        self.redliner: bool = False
        self.register: bool = False
        self.schema: bool = False
        self.student: bool = False
        self.topology: bool = False
        self.unregister: bool = False
        self.view: bool = False
        self.workspace: Optional[str] = None
        self.automation: bool = False

    def to_list(self) -> list[str]:
        result: list[str] = []
        if self.mu_create and not self.multiuser:
            raise ValueError("StartArguments.mu_create requires StartArguments.multiuser to be set to True.")

        if self.mu_open and not self.multiuser:
            raise ValueError("StartArguments.mu_open requires StartArguments.multiuser to be set to True.")

        if self.plus and not self.view:
            raise ValueError("StartArgemtns.plus requires StartArguments.view to be set to True.")

        if self.filename is not None:
            result.append(self.filename)
        if self.cable:
            result.append("/cable")
        if self.compare_config_file is not None:
            result.append("/compareconfigfile")
            result.append(self.compare_config_file)
        if self.compare_new_file is not None:
            result.append("/comparenewfile")
            result.append(self.compare_new_file)
        if self.compare_old_file is not None:
            result.append("/compareoldfile")
            result.append(self.compare_old_file)
        if self.dbe:
            result.append("/dbe")
        if self.demo:
            result.append("/demo")
        if self.dist_design:
            result.append("/distdesign")
        if self.economy:
            result.append("/economy")
        if self.fluid:
            result.append("/fluid")
        if self.formboard:
            result.append("/formboard")
        if self.functional_design:
            result.append("/functionaldesign")
        if self.inprocreg_only:
            result.append("/inprocregonly")
        if self.level:
            result.append("/level")
        if self.logic:
            result.append("/logic")
        if self.mu_create:
            result.append("/mucreate")
        if self.multiuser:
            result.append("/multiuser")
        if self.mu_open:
            result.append("/muopen")
            if isinstance(self.mu_open, str):
                result.append(self.mu_open)

        if self.new:
            result.append("/new")
        if self.no_cgm:
            result.append("/nocgm")
        if self.no_dbe:
            result.append("/nodbe")
        if self.no_export_exf:
            result.append("/noexportexf")
        if self.no_import_ruplan:
            result.append("/noimportruplan")
        if self.no_import_step:
            result.append("/noimportstep")
        if self.no_mil_standard:
            result.append("/nomilstandard")
        if self.no_panel:
            result.append("/nopanel")
        if self.no_pdf:
            result.append("/nopdf")
        if self.no_plugin:
            result.append("/noplugin")
        if self.no_splash:
            result.append("/nosplash")
        if self.no_wire:
            result.append("/nowire")
        if self.no_xvl:
            result.append("/noxvl")
        if self.plus:
            result.append("/plus")
        if self.redliner:
            result.append("/redliner")
        if self.register:
            result.append("/register")
        if self.schema:
            result.append("/schema")
        if self.student:
            result.append("/student")
        if self.topology:
            result.append("/topology")
        if self.unregister:
            result.append("/unregister")
        if self.view:
            result.append("/view")
        if self.workspace is not None:
            result.append("/Workspace=" + self.workspace)
        if self.automation:
            result.append("/automation")

        return result


def start(
    args: Union[StartArguments, list[str], None] = None,
    *,
    executable: Optional[str] = None,
    wait_for_com: bool = True,
    timeout_seconds: Optional[float] = None,
    keep_alive: bool = True,
) -> subprocess.Popen[bytes]:
    """
    Starts an E3.series process. This function returns not before the CT.Application or CT.DBEApplication COM object is ready.

    :param args: The commandline arguments for E3.series.
    :param executable: Path to the E3.series.exe. If this argument is None the latest registered E3.series.exe is executed.
    :param timeout_seconds: Timout in seconds to wait for the initialization event.
    :param keep_alive: If set to False, the E3 Process will be terminated with the script process.
    """
    if timeout_seconds is not None and timeout_seconds <= 0:
        raise ValueError("timeout_seconds must be None or positive.")

    if executable is None:
        installed_tlbs = get_registered_typelibs()
        if len(installed_tlbs) == 0:
            raise Exception("E3.series was not installed properly. If you want to start a specific E3.series, provide a executable path.")

        _, tlb_path = installed_tlbs[-1]
        tlb_directory = os.path.dirname(tlb_path)
        executable = os.path.join(tlb_directory, "E3.series.exe")

    if not os.path.isfile(executable):
        raise FileNotFoundError(f'The file "{executable}" does not exist.')

    if os.path.basename(executable).lower() != "e3.series.exe":
        raise Exception("The executable name must be 'E3.series.exe'.")

    if isinstance(args, StartArguments):
        arg_list = args.to_list()
    elif isinstance(args, list):
        arg_list = args
    elif args is None:
        arg_list = []
    elif isinstance(args, str):
        raise ValueError("Invalid value for parameter args. Plase provide a list of strings instead of a single string.")
    else:
        raise ValueError("Invalid value for parameter args.")

    if wait_for_com:
        if any((arg.startswith("/startup_event=") for arg in arg_list)):
            raise ValueError("The /startup_event argument may not be provided, when starting with wait_for_com=True.")
        event_name = str(uuid.uuid4())
        arg_list.append(f"/startup_event={event_name}")

    popen_args = [executable] + arg_list

    try:
        if wait_for_com:
            event_handle = win32event.CreateEvent(None, True, False, event_name)
            win_timeout = int(1e3 * timeout_seconds) if timeout_seconds is not None else win32event.INFINITE

        creationflags = 0
        if keep_alive:
            creationflags |= subprocess.CREATE_BREAKAWAY_FROM_JOB

        process = subprocess.Popen(popen_args, creationflags=creationflags)

        if wait_for_com:
            process_handle = int(getattr(process, "_handle"))
            event_result = win32event.WaitForMultipleObjects([event_handle, process_handle], False, win_timeout)
            if event_result == win32event.WAIT_TIMEOUT:
                raise TimeoutError("The start E3.series was aborted due to the given timeout.")

    finally:
        if wait_for_com:
            win32api.CloseHandle(event_handle)

    return process


def _get_monikers_from_rot(moniker_pattern: str, pids: Optional[list[int]] = None) -> list[tuple[Any, int]]:
    result: list[tuple[object, int]] = []

    context = pythoncom.CreateBindCtx()
    rot = pythoncom.GetRunningObjectTable()
    for moniker in rot.EnumRunning():  # type: ignore
        moniker_name = moniker.GetDisplayName(context, None)

        match = re.match(moniker_pattern, moniker_name)
        if match:
            pid = int(match.group(1))
            if pids is None or pid in pids:  # type: ignore
                result.append((rot.GetObject(moniker), pid))

    return result


def _get_app_monikers_from_rot(pids: Optional[list[int]] = None) -> list[tuple[Any, int]]:
    return _get_monikers_from_rot(r"!E3Application:(\d+)", pids)


def _get_dbe_monikers_from_rot(pids: Optional[list[int]] = None) -> list[tuple[Any, int]]:
    return _get_monikers_from_rot(r"!E3DbeApplicationFactory:(\d+)", pids)


def get_running_e3s() -> list[int]:
    """
    Returns the process ids of all running processes with an application object.
    """
    monikers = _get_app_monikers_from_rot()
    return [pid for _, pid in monikers]


def get_running_dbes() -> list[int]:
    """
    Returns the ids of all running processes with an dbe application object.
    """
    monikers = _get_dbe_monikers_from_rot()
    return [pid for _, pid in monikers]


def _check_dispatch_cast(pid: int, raw_obj: Any) -> None:
    if not hasattr(raw_obj, "PutInfo"):
        process_description = f"pid: {pid}"
        try:
            executable_path = get_executable_from_pid(pid)
            process_description += f', path: "{executable_path}"'

            version = get_version_from_executable(executable_path)
            process_description += f", version: {version}"
        except Exception:
            pass  # This only improves the error message, so if we can't get the path or version we just show the pid.

        raise RuntimeError(
            f'The COM type library for running E3.series ({process_description}) is not registered. Register the library by executing E3.series as administrator with the command line argument "/register" or reinstalling.'
        )


def _dispatch_with_errorhandling(query: Any, resultCLSID: str) -> Any:
    try:
        return win32com.client.Dispatch(query, resultCLSID=resultCLSID)
    except AttributeError as e:
        if e.name == "CLSIDToClassMap":  # type: ignore
            mod_name: str = getattr(e, "_obj").__name__
            mod_name_parts = mod_name.split(".")
            if len(mod_name_parts) == 3:
                folder_name = mod_name_parts[2]
                gen_path: str = win32com.client.gencache.GetGeneratePath()  # type: ignore
                folder_path = os.path.join(gen_path, folder_name)
                shutil.rmtree(folder_path)
                del sys.modules[mod_name]
                try:
                    return win32com.client.Dispatch(query, resultCLSID=resultCLSID)
                except AttributeError:
                    pass

    raise RuntimeError(f'Connection to "{resultCLSID}" failed.')


def _raw_connect_app(pid: int) -> Any:
    """
    Returns a raw CT.APPLICATION object from the given process.
    """
    if pid <= 0:
        raise ValueError(f"Invalid process id {pid}.")

    if not psutil.pid_exists(pid):
        raise ValueError(f"Process with pid {pid} does not exist or has already terminated.")

    objects = _get_app_monikers_from_rot([pid])
    if len(objects) == 0:
        raise Exception(f"Could not find an IApplicationInterface for pid {pid} in the RunningObjectTable.")

    iunknown, _ = objects[0]
    query = iunknown.QueryInterface(pythoncom.IID_IDispatch)
    raw_app = _dispatch_with_errorhandling(query, resultCLSID="CT.Application")
    _check_dispatch_cast(pid, raw_app)
    return raw_app


def _raw_connect_dbe(pid: int) -> Any:
    if pid <= 0:
        raise ValueError(f"Invalid process id {pid}.")

    if not psutil.pid_exists(pid):
        raise ValueError(f"Process with pid {pid} does not exist or has already terminated.")

    objects = _get_dbe_monikers_from_rot([pid])
    if len(objects) == 0:
        raise Exception(f"Could not find an IDbeApplicationInterface for pid {pid} in the RunningObjectTable.")

    iunknown, _ = objects[0]
    factory = iunknown.QueryInterface(pythoncom.IID_IClassFactory)
    instance = factory.CreateInstance(None, pythoncom.IID_IDispatch)
    del factory

    raw_dbe = _dispatch_with_errorhandling(instance, resultCLSID="CT.DbeApplication")
    _check_dispatch_cast(pid, raw_dbe)
    return raw_dbe


def _get_default_app() -> Optional[int]:
    running_e3_pids = get_running_e3s()
    if len(running_e3_pids) == 0:
        return None

    current_process = psutil.Process()
    parent_process_pids: list[int] = [process.pid for process in current_process.parents()]

    for parent_process_pid in parent_process_pids:
        if parent_process_pid in running_e3_pids:
            return parent_process_pid

    return running_e3_pids[0] if len(running_e3_pids) > 0 else None


def _get_default_dbe() -> Optional[int]:
    running_e3_pids = get_running_dbes()
    if len(running_e3_pids) == 0:
        return None

    current_process = psutil.Process()
    parent_process_pids: list[int] = [process.pid for process in current_process.parents()]

    for parent_process_pid in parent_process_pids:
        if parent_process_pid in running_e3_pids:
            return parent_process_pid

    return running_e3_pids[0] if len(running_e3_pids) > 0 else None


def _variant_to_dict(variant: Any, zeroBasedEntries: bool = True) -> dict[Any, tuple[Any, ...]]:
    result = {}
    keys = variant.keys()
    items = variant.Items()
    if len(keys) != len(items):
        raise AttributeError("Failed converting into dict!")
    if zeroBasedEntries:
        for key, item in zip(keys, items):
            result[key] = item
    else:
        for key, item in zip(keys, items):
            result[key] = item[1:] if type(item) is tuple and len(item) > 0 else tuple()
    return result
